// extac.cpp, v1.01 2013/09/26
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts TArc1.00 & TArc1.10 (*.tac) archives.

#include "as-util.h"
#include "blowfish.h"
#include "zlib.h"

struct TACHDR {
  uint8_t  signature[8]; // "TArc1.00"
  uint32_t unknown1;
  uint32_t unknown2;
  uint32_t unknown3;
  uint32_t entry_count;
  uint32_t bucket_count;
  uint32_t toc_length;
  uint32_t seed;
};

struct TACBUCKET {
  uint16_t hash16;
  uint16_t entry_count;
  uint32_t start_index;
};

struct TACENTRY {
  uint64_t hash48;
  bool     compressed;
  uint32_t original_length;
  uint32_t offset;
  uint32_t length;
};

void decrypt(uint8_t*      buff, 
             uint32_t      len, 
             const string& key)
{
  Blowfish bf;        
  bf.Set_Key((uint8_t*)key.c_str(), key.length());
  bf.Decrypt(buff, (len / 8) * 8);
}

struct TSVHDR {
  uint8_t  signature[4]; // "TSV"
  uint32_t original_length;
  uint32_t unknown1;
  uint32_t unknown2;
  uint32_t unknown3;
  uint32_t unknown4;
};

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "extac v1.01, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <input.tac>\n", argv[0]);
    return -1;
  }

  string in_filename(argv[1]);
  string prefix = as::get_file_prefix(in_filename, true);

  int fd = as::open_or_die(in_filename, O_RDONLY | O_BINARY);

  TACHDR hdr;
  read(fd, &hdr, sizeof(hdr));

  uint32_t hdr_len = sizeof(hdr);

  if (memcmp(&hdr, "TArc1.00", 8)) {
    hdr_len = 44;
  }

  lseek(fd, hdr_len, SEEK_SET);

  uint32_t toc_len  = hdr.toc_length;
  uint8_t* toc_buff = new uint8_t[toc_len];
  read(fd, toc_buff, toc_len);
  decrypt(toc_buff, toc_len, "TLibArchiveData");

  // "big enough"
  uLongf   out_toc_len  = 1024 * 1024 * 10;
  auto     out_toc_buff = new uint8_t[out_toc_len];
  int rc = uncompress(out_toc_buff, &out_toc_len, toc_buff, toc_len);

  auto buckets = (TACBUCKET*) out_toc_buff;
  auto  entries = (TACENTRY*) (buckets + hdr.bucket_count);

  uint32_t data_base = hdr_len + hdr.toc_length;
  uint32_t n         = 0;

  for (uint32_t i = 0; i < hdr.bucket_count; i++) {
    auto& bucket = buckets[i];    

    for (uint32_t j = 0; j < bucket.entry_count; j++) {
      uint32_t entry_index = bucket.start_index + j;
      auto&    entry       = entries[entry_index];

      uint64_t hash = (entry.hash48 << 16) | bucket.hash16;

      uint32_t len  = entry.length;
      auto     buff = new uint8_t[len];
      lseek(fd, data_base + entry.offset, SEEK_SET);
      read(fd, buff, len);

      if (entry.compressed) {
        uLongf out_len  = entry.original_length;
        auto   out_buff = new uint8_t[out_len];
        uncompress(out_buff, &out_len, buff, len);

        delete [] buff;

        len  = out_len;
        buff = out_buff;
      }      

      // Hm ... this may just be a coincidence in the archives I looked at.
      if (!entry.compressed) {
        string key = as::stringf("%0I64u_tlib_secure_", hash);

        uint8_t check[8];
        memcpy(check, buff, sizeof(check));
        decrypt(check, sizeof(check), key);

        uint32_t decr_len = 10240;

        if (!memcmp(check, "RIFF", 4) || !memcmp(check, "TArc", 4)) {
          decr_len = len;
        } else if (len < 10240) {
          decr_len = (len / 8) * 8;
        }

        decrypt(buff, decr_len, as::stringf("%0I64u_tlib_secure_", hash));
      }

      string ext = as::guess_file_extension(buff, len);
      if (!memcmp(buff, "TArc", 4)) ext = ".tac";

      as::write_file(prefix + as::stringf("+%05d", entry_index) + ext, 
                     buff,
                     len);

      delete [] buff;
    }
  }

  delete [] out_toc_buff;
  delete [] toc_buff;

  close(fd);

  return 0;
}